#pragma once

// #pragma message "Including the src config.h file from the binary dir."

#define PAPPSOMSPP_NAME "libpappsomspp"
#define PAPPSOMSPP_VERSION "0.11.15"
#define MASSCHROQ_VERSION "3.0.2"
/* #undef MXE */


