#pragma once

#include <QtQml>

#include "pappsomspp/export-import-config.h"
#include "pappsomspp/core/types.h"

namespace pappso {

PMSPP_LIB_DECL void registerQmlTypes(const char* uri = "pappsomspp.Enums",
                                  int versionMajor = 1,
                                  int versionMinor = 0);

PMSPP_LIB_DECL void registerEnumsToQJSEngine(QJSEngine *engine);



} // namespace pappso

